import threading
from functools import wraps


class TimeoutError(Exception):
    pass


def timeout_decorator(seconds):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            result = [TimeoutError("Function timed out after {} seconds.".format(seconds))]

            def target():
                try:
                    result[0] = func(*args, **kwargs)
                except Exception as e:
                    result[0] = e

            t = threading.Thread(target=target)
            t.start()
            t.join(seconds)
            if isinstance(result[0], BaseException):
                raise result[0]
            return result[0]

        return wrapper

    return decorator



import threading


class ThreadLocalContextStorage:
    key_value = {}

    @classmethod
    def set(cls, data):
        cls.key_value.update({threading.current_thread(): data})

    @classmethod
    def get(cls):
        cur = threading.current_thread()
        if cur not in cls.key_value:
            cls.key_value[cur] = {}
        return cls.key_value[cur]

    @classmethod
    def clear_current_thread(cls):
        del cls.key_value[threading.current_thread()]

    @classmethod
    def clear_all_thread(cls):
        cls.key_value.clear()
        cls.key_value = {}


def get_thread_context():
    if not ThreadLocalContextStorage.get():
        ThreadLocalContextStorage.set({})
    return ThreadLocalContextStorage.get()


def set_thread_context(context: dict = {}):
    ThreadLocalContextStorage.set(context)


if __name__ == '__main__':
    set_thread_context(1)
    print(get_thread_context())



from chat_agent.config import config_helper

CACHE_CHAT_LOG_EXPIRE_KEY = 'CACHE_CHAT_LOG_EXPIRE'

CACHE_CHAT_LOG_EXPIRE_DEFAULT = 3600

CACHE_CHAT_LOG_EXPIRE = config_helper.get_config_from_env(CACHE_CHAT_LOG_EXPIRE_KEY, CACHE_CHAT_LOG_EXPIRE_DEFAULT)



import json

import eventlet
from eventlet import wsgi
from flask import Flask, request, send_from_directory, session
from flask_cors import CORS
from flask_socketio import emit, SocketIO

from chat_agent import CHAT_AGENT_SECRET_KEY, CHAT_AGENT_STATIC_PATH, \
    CHAT_AGENT_SSL_CERT_FILE_FULL_PATH, CHAT_AGENT_SSL_KEY_FILE_FULL_PATH, CHAT_AGENT_HOST, CHAT_AGENT_HTTPS, \
    CHAT_AGENT_WEBSOCKET_PORT, CHAT_AGENT_WEBSOCKET_WSS_PORT
from chat_agent.handler.openai_handler import send_chat_message_with_socket_steam_response
from chat_agent.logger.logger_helper import get_logger
from chat_agent.util import random
from chat_agent.util.context import get_thread_context, set_thread_context

app = Flask(__name__, static_folder=CHAT_AGENT_STATIC_PATH)
socketio = SocketIO(app, cors_allowed_origins="*")
app.config['SECRET_KEY'] = CHAT_AGENT_SECRET_KEY
CORS(app)

logger = get_logger('websocket_app')


@app.route('/')
def index():
    sid = get_thread_context()['sid']
    ip = get_thread_context()['remote_ip']
    logger.debug('new guest, sid: {} ip: {}'.format(sid, ip))
    return send_from_directory(app.static_folder, 'websocket_index.html')


@app.route('/.<path:hidden_path>')
def ssl_check(hidden_path):
    return send_from_directory(app.static_folder, f'.{hidden_path}')


@app.before_request
def before():
    if 'sid' not in session:
        sid = random.get_random_md5()
        session['sid'] = sid
    sid = session['sid']
    get_thread_context()['sid'] = sid
    get_thread_context()['remote_ip'] = request.remote_addr


@socketio.on('connect')
def handle_connect():
    if 'sid' not in session:
        sid = random.get_random_md5()
        session['sid'] = sid
    sid = session['sid']
    get_thread_context()['sid'] = sid
    get_thread_context()['remote_ip'] = request.remote_addr
    logger.info('客户端已连接, sid: {} ip: {}'.format(sid, request.remote_addr))


@socketio.on('disconnect')
def handle_disconnect():
    sid = get_thread_context()['sid']
    logger.info('客户端已断开连接, sid: {}'.format(sid))
    set_thread_context(None)


@socketio.on('websocket_chat')
def websock_chat(msg):
    request_msg = json.loads(msg)
    if 'sid' in request_msg:
        session['sid'] = request_msg['sid']
    for response_message in send_chat_message_with_socket_steam_response(msg=request_msg['data'], chat_log=None):
        emit('websocket_chat', response_message, broadcast=True)


def start_socket_agent():
    if CHAT_AGENT_HTTPS:
        ssl_options = {
            'certfile': CHAT_AGENT_SSL_CERT_FILE_FULL_PATH,
            'keyfile': CHAT_AGENT_SSL_KEY_FILE_FULL_PATH
        }
        listener = eventlet.listen((CHAT_AGENT_HOST, CHAT_AGENT_WEBSOCKET_PORT))
        wsgi.server(eventlet.wrap_ssl(listener, **ssl_options), app)
    else:
        listener = eventlet.listen((CHAT_AGENT_HOST, CHAT_AGENT_WEBSOCKET_WSS_PORT))
        wsgi.server(listener, app)


if __name__ == '__main__':
    start_socket_agent()



import socketio

sio = socketio.Client()


@sio.event
def connect():
    print("已连接到服务器")


@sio.event
def disconnect():
    print("已断开与服务器的连接")


@sio.event
def message(data):
    print("收到消息:", data)


def main():
    sio.connect('http://localhost:8080')

    while True:
        try:
            message = input("输入消息: ")
            sio.emit('websocket_chat', message)
        except KeyboardInterrupt:
            break

    sio.disconnect()


if __name__ == '__main__':
    main()



import json

import openai
import tiktoken

from chat_agent.cache.cache_helper import get_cache, CHAT_LOG, save_cache
from chat_agent.handler import CHAT_ROUND, CHAT_TOKEN_MAX, HISTORY_TOKEN_MAX, MODEL, TEMPERATURE
from chat_agent.logger.logger_helper import get_logger
from chat_agent.util import random
from chat_agent.util.context import get_thread_context

logger = get_logger(__name__)


def num_tokens_from_messages(messages, model="gpt-3.5-turbo-0301"):
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        encoding = tiktoken.get_encoding("cl100k_base")
    if model == "gpt-3.5-turbo-0301":  # note: future models may deviate from this
        num_tokens = 0
        for message in messages:
            num_tokens += 4  # every message follows <im_start>{role/name}\n{content}<im_end>\n
            for key, value in message.items():
                num_tokens += len(encoding.encode(value))
                if key == "name":  # if there's a name, the role is omitted
                    num_tokens += -1  # role is always required and always 1 token
        num_tokens += 2  # every reply is primed with <im_start>assistant
        return num_tokens
    else:
        raise NotImplementedError(f"""num_tokens_from_messages() is not presently implemented for model {model}.
  See https://github.com/openai/openai-python/blob/main/chatml.md for information on how messages are converted to tokens.""")


def get_valid_chat_log(chat_log: []):
    sid = get_thread_context()['sid']
    valid_context_chat_log = []
    for chat in chat_log:
        if 'user' == chat['role']:
            valid_context_chat_log.append(chat)
        elif 'system' == chat['role']:
            valid_context_chat_log.append(chat)
    temp_messages = valid_context_chat_log[-CHAT_ROUND:]
    while len(temp_messages) > 1:
        token_count = num_tokens_from_messages(temp_messages)
        if token_count > HISTORY_TOKEN_MAX:
            temp_messages = temp_messages[1:]
        else:
            logger.debug('sid: {} current token count: {}'.format(sid, token_count))
            break
    return temp_messages


def send_chat_message_with_steam_response(msg: str, chat_log: [] = None):
    sid = get_thread_context()['sid']
    ip = get_thread_context()['remote_ip']
    logger.debug('receive msg, sid: {} ip: {} msg: {}'.format(sid, ip, msg))
    is_need_cache = False
    if not chat_log:
        chat_log = get_cache(CHAT_LOG, sid=sid)
        is_need_cache = True
    cur_msg = {'role': 'user', 'content': msg}
    chat_log.append(cur_msg)
    messages = get_valid_chat_log(chat_log)
    try:
        completion = openai.ChatCompletion.create(
            model=MODEL,
            messages=messages,
            temperature=TEMPERATURE,
            max_tokens=CHAT_TOKEN_MAX,
            n=1,
            stop=None,
            frequency_penalty=0,
            presence_penalty=0,
            stream=True
        )
        whole_response_msg = ''
        for chunk in completion:
            if chunk['choices'][0]['finish_reason']:
                break
            if 'role' in chunk['choices'][0]['delta']:
                continue
            chunk_text = chunk['choices'][0]['delta']['content']
            if chunk_text:
                whole_response_msg += chunk_text
                yield chunk_text
            else:
                break
        logger.debug('finish chat, sid: {} whole_response: {}'.format(sid, whole_response_msg))
        chat_log.append({
            'role': 'system', 'content': whole_response_msg
        })
        if is_need_cache:
            save_cache(CHAT_LOG, sid=sid, data=chat_log)
    except Exception as e:
        logger.error('send message has exception, msg: ' + e)


def send_chat_message_with_socket_steam_response(msg: str, chat_log: [] = None):
    msg_id = random.get_random_md5()
    index = 0
    sid = get_thread_context()['sid']
    for response in send_chat_message_with_steam_response(msg=msg, chat_log=chat_log):
        response_msg = {
            'sid': sid,
            'msg_id': msg_id,
            'index': index,
            'data': response,
            'end': False
        }
        index += 1
        yield json.dumps(response_msg)
    end_msg = {
        'sid': sid,
        'msg_id': msg_id,
        'end': True
    }
    yield json.dumps(end_msg)



import openai

from chat_agent.config import config_helper

# open api
ORGANIZATION_KEY = 'OPENAI_ORGANIZATION_KEY'
API_KEY = 'OPENAI_API_KEY'
CHAT_ROUND_KEY = 'OPENAI_CHAT_ROUND'
CHAT_TOKEN_MAX_KEY = 'OPENAI_CHAT_TOKEN_MAX'
HISTORY_TOKEN_MAX_KEY = 'OPENAI_HISTORY_TOKEN_MAX'
MODEL_KEY = 'OPENAI_MODEL'
TEMPERATURE_KEY = 'OPENAI_TEMPERATURE'

CHAT_TOKEN_MAX_DEFAULT = 2048
HISTORY_TOKEN_MAX_DEFAULT = 2048
ROUND_DEFAULT = 3
TEMPERATURE_DEFAULT = 0.7
MODEL_DEFAULT = 'gpt-3.5-turbo'

openai.organization = config_helper.get_config_from_env(ORGANIZATION_KEY)
openai.api_key = config_helper.get_config_from_env(API_KEY)
CHAT_ROUND = config_helper.get_config_from_env(CHAT_ROUND_KEY, ROUND_DEFAULT)
CHAT_TOKEN_MAX = config_helper.get_config_from_env(CHAT_TOKEN_MAX_KEY, CHAT_TOKEN_MAX_DEFAULT)
HISTORY_TOKEN_MAX = config_helper.get_config_from_env(HISTORY_TOKEN_MAX_KEY, HISTORY_TOKEN_MAX_DEFAULT)
MODEL = config_helper.get_config_from_env(MODEL_KEY, MODEL_DEFAULT)
TEMPERATURE = config_helper.get_config_from_env(TEMPERATURE_KEY, TEMPERATURE_DEFAULT)



import logging
import os.path
from logging import handlers

from chat_agent.util import file


class Logger(object):
    level_relations = {
        'debug': logging.DEBUG,
        'info': logging.INFO,
        'warning': logging.WARNING,
        'error': logging.ERROR,
        'crit': logging.CRITICAL
    }

    def __init__(self, filename, level='debug', when='D', back_count=3,
                 fmt='%(asctime)s - %(pathname)s[line:%(lineno)d] - %(levelname)s: %(message)s'):
        self.logger = logging.getLogger(filename)
        format_str = logging.Formatter(fmt)
        self.logger.setLevel(self.level_relations.get(level))
        sh = logging.StreamHandler()
        sh.setFormatter(format_str)
        th = handlers.TimedRotatingFileHandler(filename=filename, when=when, backupCount=back_count,
                                               encoding='utf-8')
        th.setFormatter(format_str)
        self.logger.addHandler(sh)
        self.logger.addHandler(th)


def get_logger(name: str):
    exec_path = file.get_exec_path()
    path = '{}/logs/'.format(exec_path)
    filename = '{}{}.log'.format(path, name)
    if not os.path.exists(path):
        os.makedirs(path)
    return Logger(filename=filename).logger


if __name__ == '__main__':
    logger = get_logger('test')
    logger.debug('debug')
    logger.info('info')
    logger.warning('警告')
    logger.error('报错')
    logger.critical('严重')



import os
import sys

import eventlet
from flask import Flask, Response, request, send_from_directory, session
from flask_cors import CORS
from flask_socketio import send, SocketIO

from chat_agent import CHAT_AGENT_PORT, CHAT_AGENT_HOST, CHAT_AGENT_SECRET_KEY, CHAT_AGENT_HTTPS, \
    CHAT_AGENT_SSL_CERT_FILE_FULL_PATH, CHAT_AGENT_SSL_KEY_FILE_FULL_PATH, CHAT_AGENT_HTTPS_PORT, \
    CHAT_AGENT_STATIC_PATH, CHAT_AGENT_TIMEOUT
from chat_agent.handler.openai_handler import send_chat_message_with_steam_response
from chat_agent.logger.logger_helper import get_logger
from chat_agent.util import random
from chat_agent.util.context import get_thread_context
from util.timeout import TimeoutError, timeout_decorator

app = Flask(__name__, static_folder=CHAT_AGENT_STATIC_PATH)
socketio = SocketIO(app)
app.config['SECRET_KEY'] = CHAT_AGENT_SECRET_KEY
CORS(app)

logger = get_logger('app')


@app.route('/')
def index():
    sid = get_thread_context()['sid']
    ip = get_thread_context()['remote_ip']
    logger.debug('new guest, sid: {} ip: {}'.format(sid, ip))
    return send_from_directory(app.static_folder, 'index.html')


@app.route('/chat')
@timeout_decorator(CHAT_AGENT_TIMEOUT)
def chat():
    try:
        msg = request.args.get('msg', '')
        chat_log = request.args.get('chat_log', [])
        return Response(send_chat_message_with_steam_response(msg=msg, chat_log=chat_log), content_type='text/plain')
    except TimeoutError:
        return Response('Timeout', content_type='text/plain'), 408


@app.route('/.<path:hidden_path>')
def ssl_check(hidden_path):
    return send_from_directory(app.static_folder, f'.{hidden_path}')


@app.before_request
def before():
    if 'sid' not in session:
        sid = random.get_random_md5()
        session['sid'] = sid
    sid = session['sid']
    get_thread_context()['sid'] = sid
    get_thread_context()['remote_ip'] = request.remote_addr


@socketio.on('websock_chat')
def websock_chat():
    try:
        msg = request.args.get('msg', '')
        chat_log = request.args.get('chat_log', [])
        for message in send_chat_message_with_steam_response(msg=msg, chat_log=chat_log):
            send(message)
    except TimeoutError:
        return Response('Timeout', content_type='text/plain'), 408


def start_agent():
    path = os.path.dirname(sys.argv[0])
    logger.debug('cur static path is: ' + path + '/' + CHAT_AGENT_STATIC_PATH)
    if CHAT_AGENT_HTTPS:
        app.run(
            host=CHAT_AGENT_HOST,
            port=CHAT_AGENT_HTTPS_PORT,
            ssl_context=(CHAT_AGENT_SSL_CERT_FILE_FULL_PATH, CHAT_AGENT_SSL_KEY_FILE_FULL_PATH)
        )
    else:
        app.run(
            host=CHAT_AGENT_HOST,
            port=CHAT_AGENT_PORT
        )


def start_socket_agent():
    eventlet.wsgi.server(eventlet.listen(('', CHAT_AGENT_PORT)), app)


if __name__ == '__main__':
    start_socket_agent()



import os


def get_config_from_env(config_key: str, default=None):
    if os.getenv(config_key):
        return os.getenv(config_key)
    else:
        return default


def get_boolean_config_from_env(config_key: str, default=None):
    val = get_config_from_env(config_key, default)
    if (isinstance(val, bool) and val) or 'True' == val or 'true' == val:
        return True
    else:
        return False



import hashlib
import random
import time

alphabet = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789~!@#$%^&*()_+[]{};,./<>?'


def get_random_md5(num: int = 64):
    random.seed(time.time())
    r = random.choices(alphabet, k=num)
    m = hashlib.md5()
    m.update(''.join(r).encode(encoding='UTF-8'))
    return m.hexdigest()



from setuptools import setup, find_packages

setup(
    name="chat_agent",
    version="0.1.6.1",
    packages=find_packages(),
    zip_sage=False,
    include_package_data=True,
    install_requires=[
        'setuptools~=63.4.1',
        'flask~=2.2.2',
        'flask_cors~=3.0.10',
        'openai~=0.27.2',
        'tiktoken~=0.3.2',
        'bidict~=0.22.1',
        'flask-socketio~=5.3.3',
        'python-engineio~=4.4.0',
        'python-socketio~=5.8.0',
        'python-engineio~=4.4.0',
        'python-socketio~=5.8.0',
        'gevent==22.10.2',
        'gevent-websocket==0.10.1',
        'eventlet==0.33.3'
    ],
    author="WhaleZarathustra",
    author_email="whale.zarathustra@gmail.com",
    description="A Simple Chat Agent, For OpenAI ChatGPT",
    license="MIT",
    keywords="ChatAgent",
    url="https://github.com/whalezarathustra/chat-agent",
)



import os

from chat_agent.config import config_helper

# agent
CHAT_AGENT_HOST_KEY = 'CHAT_AGENT_HOST'
CHAT_AGENT_PORT_KEY = 'CHAT_AGENT_PORT'
CHAT_AGENT_HTTPS_PORT_KEY = 'CHAT_AGENT_HTTPS_PORT'
CHAT_AGENT_WEBSOCKET_PORT_KEY = 'CHAT_AGENT_WEBSOCKET_PORT'
CHAT_AGENT_WEBSOCKET_WSS_PORT_KEY = 'CHAT_AGENT_WEBSOCKET_WSS_PORT'
CHAT_AGENT_SECRET_KEY = 'CHAT_AGENT_SECRET_KEY'
CHAT_AGENT_HTTPS_KEY = 'CHAT_AGENT_HTTPS'
CHAT_AGENT_SSL_CERT_FILE_FULL_PATH_KEY = 'CHAT_AGENT_SSL_CERT_FILE_FULL_PATH'
CHAT_AGENT_SSL_KEY_FILE_FULL_PATH_KEY = 'CHAT_AGENT_SSL_KEY_FILE_FULL_PATH'
CHAT_AGENT_STATIC_PATH_KEY = 'CHAT_AGENT_STATIC_PATH'
CHAT_AGENT_TIMEOUT_KEY = 'CHAT_AGENT_TIMEOUT'

CHAT_AGENT_HOST_DEFAULT = '0.0.0.0'
CHAT_AGENT_PORT_DEFAULT = 8080
CHAT_AGENT_HTTPS_PORT_DEFAULT = 8443
CHAT_AGENT_WEBSOCKET_PORT_DEFAULT = 9080
CHAT_AGENT_WEBSOCKET_WSS_PORT_DEFAULT = 9443
CHAT_AGENT_SECRET_KEY_DEFAULT = os.urandom(24)
CHAT_AGENT_HTTPS_DEFAULT = True
CHAT_AGENT_SSL_CERT_FILE_FULL_PATH_DEFAULT = 'cert.pem'
CHAT_AGENT_SSL_KEY_FILE_FULL_PATH_DEFAULT = 'key.pem'
CHAT_AGENT_STATIC_PATH_DEFAULT = 'static'
CHAT_AGENT_TIMEOUT_DEFAULT = 30

CHAT_AGENT_HOST = config_helper.get_config_from_env(CHAT_AGENT_HOST_KEY, CHAT_AGENT_HOST_DEFAULT)
CHAT_AGENT_PORT = config_helper.get_config_from_env(CHAT_AGENT_PORT_KEY, CHAT_AGENT_PORT_DEFAULT)
CHAT_AGENT_HTTPS_PORT = config_helper.get_config_from_env(CHAT_AGENT_HTTPS_PORT_KEY, CHAT_AGENT_HTTPS_PORT_DEFAULT)
CHAT_AGENT_WEBSOCKET_PORT = config_helper.get_config_from_env(CHAT_AGENT_WEBSOCKET_PORT_KEY,
                                                              CHAT_AGENT_WEBSOCKET_PORT_DEFAULT)
CHAT_AGENT_WEBSOCKET_WSS_PORT = config_helper.get_config_from_env(CHAT_AGENT_WEBSOCKET_WSS_PORT_KEY,
                                                                  CHAT_AGENT_WEBSOCKET_WSS_PORT_DEFAULT)
CHAT_AGENT_SECRET_KEY = config_helper.get_config_from_env(CHAT_AGENT_SECRET_KEY, CHAT_AGENT_SECRET_KEY_DEFAULT)
CHAT_AGENT_HTTPS = config_helper.get_boolean_config_from_env(CHAT_AGENT_HTTPS_KEY, CHAT_AGENT_HTTPS_DEFAULT)
CHAT_AGENT_SSL_CERT_FILE_FULL_PATH = config_helper.get_config_from_env(CHAT_AGENT_SSL_CERT_FILE_FULL_PATH_KEY,
                                                                       CHAT_AGENT_SSL_CERT_FILE_FULL_PATH_DEFAULT)
CHAT_AGENT_SSL_KEY_FILE_FULL_PATH = config_helper.get_config_from_env(CHAT_AGENT_SSL_KEY_FILE_FULL_PATH_KEY,
                                                                      CHAT_AGENT_SSL_KEY_FILE_FULL_PATH_DEFAULT)
CHAT_AGENT_STATIC_PATH = config_helper.get_config_from_env(CHAT_AGENT_STATIC_PATH_KEY, CHAT_AGENT_STATIC_PATH_DEFAULT)
CHAT_AGENT_TIMEOUT = config_helper.get_config_from_env(CHAT_AGENT_TIMEOUT_KEY, CHAT_AGENT_TIMEOUT_DEFAULT)



from threading import Timer

from chat_agent.cache import CACHE_CHAT_LOG_EXPIRE
from chat_agent.util.time import get_timestamp

CHAT_LOG = 'chat_log'

memory_cache = {
    CHAT_LOG: {}
}


def get_cache(key: str, sid: str):
    if sid not in memory_cache[key]:
        return []
    return memory_cache[key][sid]['data']


def save_cache(key: str, sid: str, data):
    cache_data = {
        'timestamp': get_timestamp(),
        'data': data
    }
    memory_cache[key][sid] = cache_data


def clean_expire_data():
    cur_timestamp = get_timestamp()
    for sid in list(memory_cache[CHAT_LOG].keys()):
        if memory_cache[CHAT_LOG][sid]['timestamp'] + CACHE_CHAT_LOG_EXPIRE < cur_timestamp:
            del memory_cache[CHAT_LOG][sid]


def clean():
    clean_expire_data()
    Timer(1, clean).start()


clean()
